@@ -1,9 +1,12 @@
 import asyncio
 import contextlib
 import os
+import platform
+import signal
 import socket
 import ssl
-from io import StringIO
+import subprocess
+import sys
 from unittest import mock
 from uuid import uuid4
 
@@ -39,6 +42,16 @@ skip_if_no_unix_socks = pytest.mark.skipif(
 )
 del _has_unix_domain_socks, _abstract_path_failed
 
+HAS_IPV6 = socket.has_ipv6
+if HAS_IPV6:
+    # The socket.has_ipv6 flag may be True if Python was built with IPv6
+    # support, but the target system still may not have it.
+    # So let's ensure that we really have IPv6 support.
+    try:
+        socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
+    except OSError:
+        HAS_IPV6 = False
+
 
 # tokio event loop does not allow to override attributes
 def skip_if_no_dict(loop):
@@ -46,16 +59,26 @@ def skip_if_no_dict(loop):
         pytest.skip("can not override loop attributes")
 
 
+def skip_if_on_windows():
+    if platform.system() == "Windows":
+        pytest.skip("the test is not valid for Windows")
+
+
+def stopper(loop):
+    def f(*args):
+        loop.call_later(0.001, loop.stop)
+    return f
+
+
 def test_run_app_http(loop, mocker):
     skip_if_no_dict(loop)
 
     mocker.spy(loop, 'create_server')
-    loop.call_later(0.05, loop.stop)
 
     app = web.Application()
     mocker.spy(app, 'startup')
 
-    web.run_app(app, loop=loop, print=lambda *args: None)
+    web.run_app(app, loop=loop, print=stopper(loop))
 
     assert not loop.is_closed()
     loop.create_server.assert_called_with(mock.ANY, '0.0.0.0', 8080,
@@ -67,14 +90,13 @@ def test_run_app_close_loop(loop, mocker):
     skip_if_no_dict(loop)
 
     mocker.spy(loop, 'create_server')
-    loop.call_later(0.05, loop.stop)
 
     asyncio.set_event_loop(loop)
 
     app = web.Application()
     mocker.spy(app, 'startup')
 
-    web.run_app(app, print=lambda *args: None)
+    web.run_app(app, print=stopper(loop))
 
     assert loop.is_closed()
     loop.create_server.assert_called_with(mock.ANY, '0.0.0.0', 8080,
@@ -202,7 +224,7 @@ def test_run_app_mixed_bindings(mocker, run_app_kwargs, expected_server_calls,
     loop = mocker.MagicMock()
     mocker.patch('asyncio.gather')
 
-    web.run_app(app, loop=loop, print=lambda *args: None, **run_app_kwargs)
+    web.run_app(app, loop=loop, print=None, **run_app_kwargs)
 
     assert loop.create_unix_server.mock_calls == expected_unix_server_calls
     assert loop.create_server.mock_calls == expected_server_calls
@@ -212,13 +234,12 @@ def test_run_app_http_access_format(loop, mocker):
     skip_if_no_dict(loop)
 
     mocker.spy(loop, 'create_server')
-    loop.call_later(0.05, loop.stop)
 
     app = web.Application()
     mocker.spy(app, 'startup')
 
     web.run_app(app, loop=loop,
-                print=lambda *args: None, access_log_format='%a')
+                print=stopper(loop), access_log_format='%a')
 
     assert not loop.is_closed()
     cs = loop.create_server
@@ -231,7 +252,6 @@ def test_run_app_https(loop, mocker):
     skip_if_no_dict(loop)
 
     mocker.spy(loop, 'create_server')
-    loop.call_later(0.05, loop.stop)
 
     app = web.Application()
     mocker.spy(app, 'startup')
@@ -239,7 +259,7 @@ def test_run_app_https(loop, mocker):
     ssl_context = ssl.create_default_context()
 
     web.run_app(app, loop=loop,
-                ssl_context=ssl_context, print=lambda *args: None)
+                ssl_context=ssl_context, print=stopper(loop))
 
     assert not loop.is_closed()
     loop.create_server.assert_called_with(mock.ANY, '0.0.0.0', 8443,
@@ -251,16 +271,15 @@ def test_run_app_nondefault_host_port(loop, unused_port, mocker):
     skip_if_no_dict(loop)
 
     port = unused_port()
-    host = 'localhost'
+    host = '127.0.0.1'
 
     mocker.spy(loop, 'create_server')
-    loop.call_later(0.05, loop.stop)
 
     app = web.Application()
     mocker.spy(app, 'startup')
 
     web.run_app(app, loop=loop,
-                host=host, port=port, print=lambda *args: None)
+                host=host, port=port, print=stopper(loop))
 
     assert not loop.is_closed()
     loop.create_server.assert_called_with(mock.ANY, host, port,
@@ -272,12 +291,11 @@ def test_run_app_custom_backlog(loop, mocker):
     skip_if_no_dict(loop)
 
     mocker.spy(loop, 'create_server')
-    loop.call_later(0.05, loop.stop)
 
     app = web.Application()
     mocker.spy(app, 'startup')
 
-    web.run_app(app, loop=loop, backlog=10, print=lambda *args: None)
+    web.run_app(app, loop=loop, backlog=10, print=stopper(loop))
 
     assert not loop.is_closed()
     loop.create_server.assert_called_with(mock.ANY, '0.0.0.0', 8080,
@@ -290,20 +308,20 @@ def test_run_app_http_unix_socket(loop, mocker, shorttmpdir):
     skip_if_no_dict(loop)
 
     mocker.spy(loop, 'create_unix_server')
-    loop.call_later(0.05, loop.stop)
 
     app = web.Application()
     mocker.spy(app, 'startup')
 
     sock_path = str(shorttmpdir.join('socket.sock'))
-    printed = StringIO()
-    web.run_app(app, loop=loop, path=sock_path, print=printed.write)
+    printer = mock.Mock(wraps=stopper(loop))
+    web.run_app(app, loop=loop, path=sock_path,
+                print=printer)
 
     assert not loop.is_closed()
     loop.create_unix_server.assert_called_with(mock.ANY, sock_path,
                                                ssl=None, backlog=128)
     app.startup.assert_called_once_with()
-    assert "http://unix:{}:".format(sock_path) in printed.getvalue()
+    assert "http://unix:{}:".format(sock_path) in printer.call_args[0][0]
 
 
 @skip_if_no_unix_socks
@@ -311,22 +329,21 @@ def test_run_app_https_unix_socket(loop, mocker, shorttmpdir):
     skip_if_no_dict(loop)
 
     mocker.spy(loop, 'create_unix_server')
-    loop.call_later(0.05, loop.stop)
 
     app = web.Application()
     mocker.spy(app, 'startup')
 
     sock_path = str(shorttmpdir.join('socket.sock'))
-    printed = StringIO()
     ssl_context = ssl.create_default_context()
+    printer = mock.Mock(wraps=stopper(loop))
     web.run_app(app, loop=loop, path=sock_path, ssl_context=ssl_context,
-                print=printed.write)
+                print=printer)
 
     assert not loop.is_closed()
     loop.create_unix_server.assert_called_with(mock.ANY, sock_path,
                                                ssl=ssl_context, backlog=128)
     app.startup.assert_called_once_with()
-    assert "https://unix:{}:".format(sock_path) in printed.getvalue()
+    assert "https://unix:{}:".format(sock_path) in printer.call_args[0][0]
 
 
 @skip_if_no_unix_socks
@@ -336,31 +353,28 @@ def test_run_app_stale_unix_socket(loop, mocker, shorttmpdir):
     """
     skip_if_no_dict(loop)
 
-    loop.call_later(0.05, loop.stop)
-
     app = web.Application()
 
     sock_path = shorttmpdir.join('socket.sock')
     sock_path_string = str(sock_path)
 
     web.run_app(app, loop=loop,
-                path=sock_path_string, print=lambda *args: None)
+                path=sock_path_string, print=stopper(loop))
     assert not loop.is_closed()
 
     if sock_path.check():
         # New app run using same socket path
         with loop_context() as loop:
             mocker.spy(loop, 'create_unix_server')
-            loop.call_later(0.05, loop.stop)
 
             app = web.Application()
 
             mocker.spy(app, 'startup')
             mocker.spy(os, 'remove')
-            printed = StringIO()
+            printer = mock.Mock(wraps=stopper(loop))
 
             web.run_app(app, loop=loop,
-                        path=sock_path_string, print=printed.write)
+                        path=sock_path_string, print=printer)
             os.remove.assert_called_with(sock_path_string)
             loop.create_unix_server.assert_called_with(
                 mock.ANY,
@@ -369,8 +383,8 @@ def test_run_app_stale_unix_socket(loop, mocker, shorttmpdir):
                 backlog=128
             )
             app.startup.assert_called_once_with()
-            assert "http://unix:{}:".format(sock_path) in \
-                   printed.getvalue()
+            assert ("http://unix:{}:".format(sock_path)
+                    in printer.call_args[0][0])
 
 
 @skip_if_no_unix_socks
@@ -378,24 +392,21 @@ def test_run_app_stale_unix_socket(loop, mocker, shorttmpdir):
 def test_run_app_abstract_linux_socket(loop, mocker):
     sock_path = b"\x00" + uuid4().hex.encode('ascii')
 
-    loop.call_later(0.05, loop.stop)
     app = web.Application()
     web.run_app(
         app, path=sock_path.decode('ascii', 'ignore'), loop=loop,
-        print=lambda *args: None)
+        print=stopper(loop))
 
     # New app run using same socket path
     with loop_context() as loop:
         mocker.spy(loop, 'create_unix_server')
-        loop.call_later(0.05, loop.stop)
 
         app = web.Application()
 
         mocker.spy(app, 'startup')
         mocker.spy(os, 'remove')
-        printed = StringIO()
 
-        web.run_app(app, path=sock_path, print=printed.write, loop=loop)
+        web.run_app(app, path=sock_path, print=stopper(loop), loop=loop)
 
         # Abstract paths don't exist on the file system, so no attempt should
         # be made to remove.
@@ -420,7 +431,7 @@ def test_run_app_existing_file_conflict(loop, mocker, shorttmpdir):
 
     with pytest.raises(OSError):
         web.run_app(app, loop=loop,
-                    path=sock_path_str, print=lambda *args: None)
+                    path=sock_path_str, print=mock.Mock())
 
     # No attempt should be made to remove a non-socket file
     assert mock.call([sock_path_str]) not in os.remove.mock_calls
@@ -430,7 +441,6 @@ def test_run_app_preexisting_inet_socket(loop, mocker):
     skip_if_no_dict(loop)
 
     mocker.spy(loop, 'create_server')
-    loop.call_later(0.05, loop.stop)
 
     app = web.Application()
     mocker.spy(app, 'startup')
@@ -440,15 +450,40 @@ def test_run_app_preexisting_inet_socket(loop, mocker):
         sock.bind(('0.0.0.0', 0))
         _, port = sock.getsockname()
 
-        printed = StringIO()
-        web.run_app(app, loop=loop, sock=sock, print=printed.write)
+        printer = mock.Mock(wraps=stopper(loop))
+        web.run_app(app, loop=loop, sock=sock, print=printer)
 
         assert not loop.is_closed()
         loop.create_server.assert_called_with(
             mock.ANY, sock=sock, backlog=128, ssl=None
         )
         app.startup.assert_called_once_with()
-        assert "http://0.0.0.0:{}".format(port) in printed.getvalue()
+        assert "http://0.0.0.0:{}".format(port) in printer.call_args[0][0]
+
+
+@pytest.mark.skipif(not HAS_IPV6, reason="IPv6 is not available")
+def test_run_app_preexisting_inet6_socket(loop, mocker):
+    skip_if_no_dict(loop)
+
+    mocker.spy(loop, 'create_server')
+
+    app = web.Application()
+    mocker.spy(app, 'startup')
+
+    sock = socket.socket(socket.AF_INET6)
+    with contextlib.closing(sock):
+        sock.bind(('::', 0))
+        port = sock.getsockname()[1]
+
+        printer = mock.Mock(wraps=stopper(loop))
+        web.run_app(app, loop=loop, sock=sock, print=printer)
+
+        assert not loop.is_closed()
+        loop.create_server.assert_called_with(
+            mock.ANY, sock=sock, backlog=128, ssl=None
+        )
+        app.startup.assert_called_once_with()
+        assert "http://:::{}".format(port) in printer.call_args[0][0]
 
 
 @skip_if_no_unix_socks
@@ -456,7 +491,6 @@ def test_run_app_preexisting_unix_socket(loop, mocker):
     skip_if_no_dict(loop)
 
     mocker.spy(loop, 'create_server')
-    loop.call_later(0.05, loop.stop)
 
     app = web.Application()
     mocker.spy(app, 'startup')
@@ -467,22 +501,21 @@ def test_run_app_preexisting_unix_socket(loop, mocker):
         sock.bind(sock_path)
         os.unlink(sock_path)
 
-        printed = StringIO()
-        web.run_app(app, loop=loop, sock=sock, print=printed.write)
+        printer = mock.Mock(wraps=stopper(loop))
+        web.run_app(app, loop=loop, sock=sock, print=printer)
 
         assert not loop.is_closed()
         loop.create_server.assert_called_with(
             mock.ANY, sock=sock, backlog=128, ssl=None
         )
         app.startup.assert_called_once_with()
-        assert "http://unix:{}:".format(sock_path) in printed.getvalue()
+        assert "http://unix:{}:".format(sock_path) in printer.call_args[0][0]
 
 
 def test_run_app_multiple_preexisting_sockets(loop, mocker):
     skip_if_no_dict(loop)
 
     mocker.spy(loop, 'create_server')
-    loop.call_later(0.05, loop.stop)
 
     app = web.Application()
     mocker.spy(app, 'startup')
@@ -495,13 +528,76 @@ def test_run_app_multiple_preexisting_sockets(loop, mocker):
         sock2.bind(('0.0.0.0', 0))
         _, port2 = sock2.getsockname()
 
-        printed = StringIO()
-        web.run_app(app, loop=loop, sock=(sock1, sock2), print=printed.write)
+        printer = mock.Mock(wraps=stopper(loop))
+        web.run_app(app, loop=loop, sock=(sock1, sock2), print=printer)
 
         loop.create_server.assert_has_calls([
             mock.call(mock.ANY, sock=sock1, backlog=128, ssl=None),
             mock.call(mock.ANY, sock=sock2, backlog=128, ssl=None)
         ])
         app.startup.assert_called_once_with()
-        assert "http://0.0.0.0:{}".format(port1) in printed.getvalue()
-        assert "http://0.0.0.0:{}".format(port2) in printed.getvalue()
+        assert "http://0.0.0.0:{}".format(port1) in printer.call_args[0][0]
+        assert "http://0.0.0.0:{}".format(port2) in printer.call_args[0][0]
+
+
+_script_test_signal = """
+from aiohttp import web
+
+app = web.Application()
+web.run_app(app, host=())
+"""
+
+
+def test_sigint(loop, mocker):
+    skip_if_on_windows()
+
+    proc = subprocess.Popen([sys.executable, "-u", "-c", _script_test_signal],
+                            stdout=subprocess.PIPE)
+    for line in proc.stdout:
+        if line.startswith(b"======== Running on"):
+            break
+    proc.send_signal(signal.SIGINT)
+    assert proc.wait() == 0
+
+
+def test_sigterm(loop, mocker):
+    skip_if_on_windows()
+
+    proc = subprocess.Popen([sys.executable, "-u", "-c", _script_test_signal],
+                            stdout=subprocess.PIPE)
+    for line in proc.stdout:
+        if line.startswith(b"======== Running on"):
+            break
+    proc.terminate()
+    assert proc.wait() == 0
+
+
+def test_startup_cleanup_signals(loop, mocker):
+    skip_if_no_dict(loop)
+
+    mocker.spy(loop, 'create_server')
+
+    app = web.Application()
+    mocker.spy(app, 'startup')
+    mocker.spy(app, 'cleanup')
+
+    web.run_app(app, loop=loop, host=(), print=stopper(loop))
+
+    app.startup.assert_called_once_with()
+    app.cleanup.assert_called_once_with()
+
+
+def test_startup_cleanup_signals_even_on_failure(loop, mocker):
+    skip_if_no_dict(loop)
+
+    setattr(loop, 'create_server', mock.Mock(side_effect=RuntimeError()))
+
+    app = web.Application()
+    mocker.spy(app, 'startup')
+    mocker.spy(app, 'cleanup')
+
+    with pytest.raises(RuntimeError):
+        web.run_app(app, loop=loop, print=stopper(loop))
+
+    app.startup.assert_called_once_with()
+    app.cleanup.assert_called_once_with()
